# pylint: disable=missing-docstring, invalid-name
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('helcim', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='helcimtoken',
            name='cc_name',
            field=models.CharField(
                blank=True,
                help_text='The cardholder name',
                max_length=256,
                null=True
            ),
        ),
        migrations.AddField(
            model_name='helcimtoken',
            name='cc_expiry',
            field=models.DateField(
                blank=True,
                help_text='The credit card expiry date',
                null=True
            ),
        ),
        migrations.AlterField(
            model_name='helcimtoken',
            name='django_user',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='helcim_tokens',
                to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



# pylint: disable=missing-docstring, invalid-name
import re
import sys
import warnings

import django


__version__ = '0.9.1'

# Provide DepreciationWarning for older Python versions
# Have to use sys.version while supporting Python 3.5 to enable testing
# Once Python 3.5 is dropped can switch to version_info & compare tuples
if re.match(r'^3\.5', sys.version):
    warnings.warn(
        (
            'django-helcim will stop supporting Python 3.5 once it '
            'reaches end-of-life (approximately September 2020). '
            'Ensure you have updated your Python version by then.'
        ),
        DeprecationWarning
    )
# Provide DepreciationWarning for older Django versions
if '1.11' in django.__version__:
    warnings.warn(
        (
            'django-helcim will stop supporting Django 1.11 LTS once '
            'it reaches end-of-life (approximately April 2020). '
            'Ensure you have updated your Django version by then.'
        ),
        DeprecationWarning
    )

# Django configuration details
default_app_config = 'helcim.apps.HelcimConfig'



"""Determines relevant settings for django-helcim functioning."""
from django.conf import settings as django_settings
from django.core import exceptions as django_exceptions

def _validate_helcim_js_settings(helcim_js):
    """Confirms that declared Helcim.js are in proper format."""
    if isinstance(helcim_js, dict) is False:
        message = 'HELCIM_JS_CONFIG setting must be a dictionary.'
        raise django_exceptions.ImproperlyConfigured(message)

    for _, value in helcim_js.items():
        if 'url' not in value or 'token' not in value:
            message = (
                'HELCIM_JS_CONFIG values must include both a '
                '"url" and "token" key.'
            )
            raise django_exceptions.ImproperlyConfigured(message)

def determine_helcim_settings():
    """Collects all possible django-helcim settings for easy use.

        Performs basic validation of required settings and assigns
        defaults where applicable.

        Returns:
            dict: Summary of all possible django-helcim settings.
    """
    # API AND HELCIM.JS SETTINGS
    # -------------------------------------------------------------------------
    # API Settings
    # TODO: rework these setting names to be clearly API-specific
    # HELCIM_ACCOUNT_ID to become HELCIM_API_ACCOUNT_ID
    # HELCIM_TERMINAL_ID TO BECOME HELCIM_API_TERMINAL_ID
    account_id = getattr(django_settings, 'HELCIM_ACCOUNT_ID', '')
    api_token = getattr(django_settings, 'HELCIM_API_TOKEN', '')
    api_url = getattr(
        django_settings, 'HELCIM_API_URL', 'https://secure.myhelcim.com/api/'
    )
    terminal_id = getattr(django_settings, 'HELCIM_TERMINAL_ID', '')
    api_test = getattr(django_settings, 'HELCIM_API_TEST', False)

    # Helcim.js Settings
    helcim_js = getattr(django_settings, 'HELCIM_JS_CONFIG', {})
    _validate_helcim_js_settings(helcim_js)

    # REDACTION SETTINGS
    # -------------------------------------------------------------------------
    redact_all = getattr(django_settings, 'HELCIM_REDACT_ALL', None)
    redact_cc_name = getattr(django_settings, 'HELCIM_REDACT_CC_NAME', True)
    redact_cc_number = getattr(
        django_settings, 'HELCIM_REDACT_CC_NUMBER', True
    )
    redact_cc_expiry = getattr(
        django_settings, 'HELCIM_REDACT_CC_EXPIRY', True
    )
    redact_cc_cvv = getattr(django_settings, 'HELCIM_REDACT_CC_CVV', True)
    redact_cc_type = getattr(django_settings, 'HELCIM_REDACT_CC_TYPE', True)
    redact_cc_magnetic = getattr(
        django_settings, 'HELCIM_REDACT_CC_MAGNETIC', True
    )
    redact_cc_magnetic_encrypted = getattr(
        django_settings, 'HELCIM_REDACT_CC_MAGNETIC_ENCRYPTED', True
    )
    redact_token = getattr(django_settings, 'HELCIM_REDACT_TOKEN', False)

    # TRANSACTION FUNCTIONALITY SETTINGS
    # -------------------------------------------------------------------------
    enable_transaction_capture = getattr(
        django_settings, 'HELCIM_ENABLE_TRANSACTION_CAPTURE', False
    )
    enable_transaction_refund = getattr(
        django_settings, 'HELCIM_ENABLE_TRANSACTION_REFUND', False
    )

    # TOKEN VAULT SETTINGS
    # -------------------------------------------------------------------------
    enable_token_vault = getattr(
        django_settings, 'HELCIM_ENABLE_TOKEN_VAULT', False
    )

    # ADMIN SETTINGS
    # -------------------------------------------------------------------------
    enable_admin = getattr(
        django_settings, 'HELCIM_ENABLE_ADMIN', False
    )

    # OTHER SETTINGS
    # -------------------------------------------------------------------------
    allow_anonymous = getattr(
        django_settings, 'HELCIM_ALLOW_ANONYMOUS', True
    )

    return {
        'api_url': api_url,
        'account_id': account_id,
        'api_token': api_token,
        'terminal_id': terminal_id,
        'api_test': api_test,
        'helcim_js': helcim_js,
        'redact_all': redact_all,
        'redact_cc_name': redact_cc_name,
        'redact_cc_number': redact_cc_number,
        'redact_cc_expiry': redact_cc_expiry,
        'redact_cc_cvv': redact_cc_cvv,
        'redact_cc_type': redact_cc_type,
        'redact_cc_magnetic': redact_cc_magnetic,
        'redact_cc_magnetic_encrypted': redact_cc_magnetic_encrypted,
        'redact_token': redact_token,
        'enable_transaction_capture': enable_transaction_capture,
        'enable_transaction_refund': enable_transaction_refund,
        'enable_token_vault': enable_token_vault,
        'enable_admin': enable_admin,
        'allow_anonymous': allow_anonymous,
    }

SETTINGS = determine_helcim_settings()



"""Mixins to help support Helcim API interactions."""
from calendar import monthrange
import copy
from datetime import datetime
import re

import pytz

from django.db import IntegrityError
from django.utils.safestring import mark_safe

from helcim import exceptions as helcim_exceptions
from helcim.settings import SETTINGS
from helcim.models import HelcimToken, HelcimTransaction


class ResponseMixin():
    """Methods to support handling a Helcim repsonse.

        Handles some data manipulations to prepare for saving to a
        model instance and applies relevant redactions to data.
     """
    @classmethod
    def _identify_redact_fields(cls):
        """Identifies which fields (if any) should be redacted.

            Configured using flags in the Django settings file.

            Returns:
                dict: A dictionary of fields to redact with corresponding
                    API field and variable names.
        """
        redact_fields = {
            'name': {
                'redact': None,
                'settings': 'redact_cc_name',
                'fields': [
                    {'api': 'cardHolderName', 'python': 'cc_name'},
                ]
            },
            'number': {
                'redact': None,
                'settings': 'redact_cc_number',
                'fields': [
                    {'api': 'cardNumber', 'python': 'cc_number'},
                ]
            },
            'expiry': {
                'redact': None,
                'settings': 'redact_cc_expiry',
                'fields': [
                    {'api': 'cardExpiry', 'python': 'cc_expiry'},
                    {'api': 'expiryDate', 'python': 'cc_expiry'},
                ]
            },
            'cvv': {
                'redact': None,
                'settings': 'redact_cc_cvv',
                'fields': [
                    {'api': 'cardCVV', 'python': 'cc_cvv'},
                ]
            },
            'type': {
                'redact': None,
                'settings': 'redact_cc_type',
                'fields': [
                    {'api': 'cardType', 'python': 'cc_type'},
                ]
            },
            'token': {
                'redact': None,
                'settings': 'redact_token',
                'fields': [
                    {'api': 'cardToken', 'python': 'token'},
                    {'api': 'cardF4L4', 'python': 'token_f4l4'},
                ]
            },
            'mag': {
                'redact': None,
                'settings': 'redact_cc_magnetic',
                'fields': [
                    {'api': 'cardMag', 'python': 'mag'},
                ]
            },
            'mag_enc': {
                'redact': None,
                'settings': 'redact_cc_magnetic_encrypted',
                'fields': [
                    {'api': 'cardMagEnc', 'python': 'mag_enc'},
                    {'api': 'serialNumber', 'python': 'mang_enc_serial_number'}, # pylint: disable=line-too-long
                ]
            },
        }

        # HELCIM_REDACT_ALL overrides all other settings
        if SETTINGS['redact_all'] is not None:
            if SETTINGS['redact_all'] is True:
                for key in redact_fields:
                    redact_fields[key]['redact'] = True
            else:
                for key in redact_fields:
                    redact_fields[key]['redact'] = False

        # Otherwise, assess each field individually
        else:
            for key, field in redact_fields.items():
                redact_fields[key]['redact'] = SETTINGS[field['settings']]

        return redact_fields

    @classmethod
    def _convert_expiry_to_date(cls, expiry):
        """Converts a 4 digit expiry to a datetime object.

            Parameters:
                expiry (str): the four digit representation of the
                    credit card expiry

            Returns:
                obj: the expiry as a datetime object.
        """

        year = 2000 + int(expiry[2:])
        month = int(expiry[:2])
        day = monthrange(year, month)[1]

        return datetime(year, month, day, tzinfo=pytz.timezone('UTC')).date()

    @classmethod
    def _determine_save_token_status(cls, user_decision):
        """Determines if Helcim card token should be saved.

            Parameters:
                user_decision (bool): Whether the user has requested to
                    save this token or not.

            Returns:
                bool: Whether a token should be saved.
        """
        # Check if vault is enabled in settings
        vault_enabled = SETTINGS['enable_token_vault']

        # If yes, check if user requested save
        if vault_enabled:
            return user_decision

        return False

    def _determine_user_reference(self):
        """Validates and returns appropriate user reference.

            Determines if an absent or anonymous user is permitted.
            Returns an exception if they are provided and not allowed.
        """
        # Handles anonymous/no user situations
        if SETTINGS['allow_anonymous']:
            if self.django_user is None or self.django_user.is_anonymous:
                return None
        else:
            if self.django_user is None or self.django_user.is_anonymous:
                raise helcim_exceptions.ProcessingError(
                    'Required Django user reference not provided.'
                )

        # Otherwise can just return the provided user model
        return self.django_user

    def _redact_api_data(self):
        """Redacts API data and updates redacted_response attribute."""
        if 'raw_request' in self.redacted_response:
            self.redacted_response['raw_request'] = re.sub(
                r'(accountId=.*?)(&|$)',
                r'accountId=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
            self.redacted_response['raw_request'] = re.sub(
                r'(apiToken=.*?)(&|$)',
                r'apiToken=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
            self.redacted_response['raw_request'] = re.sub(
                r'(terminalId=.*?)(&|$)',
                r'terminalId=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
        else:
            self.redacted_response['raw_request'] = None

    def _redact_field(self, api_name, python_name):
        """Redacts all information for the provided field.

            Method directly updates the redacted_response attribute.

            Parameters:
                api_name (str): The field name used by the Helcim API.
                python_name (str): The field name used by this
                    application.
        """
        # Redacts the raw_request data (if present)
        if self.redacted_response.get('raw_request', None):
            self.redacted_response['raw_request'] = re.sub(
                r'({}=.*?)(&|$)'.format(api_name),
                r'{}=REDACTED\g<2>'.format(api_name),
                self.redacted_response['raw_request']
            )

        # Redacts the raw_response data (if present)
        if self.redacted_response.get('raw_response', None):
            self.redacted_response['raw_response'] = re.sub(
                r'<{0}>.*</{0}>'.format(api_name),
                r'<{0}>REDACTED</{0}>'.format(api_name),
                self.redacted_response['raw_response']
            )

        if python_name in self.redacted_response:
            self.redacted_response[python_name] = None

    def redact_data(self):
        """Removes sensitive and identifiable data.

            By default will redact API fields and populate
            redacted_response attribute. Depending on Django settings,
            may also redact other fields in the formated and raw
            response.
        """
        # Copy the response data to the redacted file for updating
        self.redacted_response = copy.deepcopy(self.response)

        # Remove any API content
        self._redact_api_data()

        # Identify and redact any other specified fields
        fields = self._identify_redact_fields()

        for _, redact_field in fields.items():
            if redact_field['redact']:
                for field in redact_field['fields']:
                    self._redact_field(field['api'], field['python'])

        if fields['name']['redact']:
            self.response['cc_name'] = None

        if fields['expiry']['redact']:
            self.response['cc_expiry'] = None

    def create_model_arguments(self, transaction_type):
        """Creates dictionary for use as transaction model arguments.

            Takes the redacted_response data and creates a dictionary
            that can be used as the keyword argumetns for the
            HelcimTransaction model.

            Parameters:
                transaction_type (str): Transaction type for this
                    transaction.

            Returns:
                dict: dictionary of the HelcimTransaction arguments.
        """
        response = self.redacted_response

        # Format the transaction_date
        if all([
                response.get('transaction_date'),
                response.get('transaction_time')
        ]):
            date_response = datetime.combine(
                response.get('transaction_date'),
                response.get('transaction_time')
            )
        else:
            date_response = None

        # Format the credit card expiry date (if present)
        if response.get('cc_expiry', None):
            cc_expiry = self._convert_expiry_to_date(
                response['cc_expiry']
            )
        else:
            cc_expiry = None

        # Return proper user reference
        django_user = self._determine_user_reference()

        return {
            'raw_request': response.get('raw_request'),
            'raw_response': response.get('raw_response'),
            'transaction_success': response.get('transaction_success'),
            'response_message': response.get('response_message'),
            'notice': response.get('notice'),
            'date_response': date_response,
            'transaction_type': transaction_type,
            'transaction_id': response.get('transaction_id'),
            'amount': response.get('amount'),
            'currency': response.get('currency'),
            'cc_name': response.get('cc_name'),
            'cc_number': response.get('cc_number'),
            'cc_expiry': cc_expiry,
            'cc_type': response.get('cc_type'),
            'token': response.get('token'),
            'token_f4l4': response.get('token_f4l4'),
            'avs_response': response.get('avs_response'),
            'cvv_response': response.get('cvv_response'),
            'approval_code': response.get('approval_code'),
            'order_number': response.get('order_number'),
            'customer_code': response.get('customer_code'),
            'django_user': django_user,
        }

    def save_transaction(self, transaction_type):
        """Saves HelcimTransaction with redacted response details."""
        # Redacts data if not already done
        if not self.redacted_response:
            self.redact_data()

        model_dictionary = self.create_model_arguments(transaction_type)

        try:
            transaction_instance = HelcimTransaction.objects.create(
                **model_dictionary
            )
        except IntegrityError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )
        except ValueError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )

        return transaction_instance

    def save_token_to_vault(self):
        """Saves Helcim card token.

            Returns:
                obj: The HelcimToken model instance, or ``None`` (if
                    model not created).
        """
        # Confirms that the token should be saved
        if self.save_token is False:
            return None

        token = self.response.get('token', None)
        token_f4l4 = self.response.get('token_f4l4', None)
        cc_name = self.response.get('cc_name', None)
        raw_expiry = self.response.get('cc_expiry', None)
        cc_expiry = (
            self._convert_expiry_to_date(raw_expiry) if raw_expiry else None
        )

        # Ensure there is a customer code (can't use token without one)
        try:
            customer_code = self.response['customer_code']
        except KeyError:
            raise helcim_exceptions.ProcessingError(
                'Unable to save token - customer code not provided'
            )

        # Retrieve proper user reference
        django_user = self._determine_user_reference()

        if token and token_f4l4:
            token_instance, _ = HelcimToken.objects.get_or_create(
                token=token,
                token_f4l4=token_f4l4,
                cc_name=cc_name,
                cc_expiry=cc_expiry,
                cc_type=self.response.get('cc_type', None),
                customer_code=customer_code,
                django_user=django_user,
            )

            return token_instance

        # If unable to save token, return None
        return None

class HelcimJSMixin():
    """Provides Helcim.js URL and token details in the view context.

        This is a helper mixin that allows you to declare your Helcim.js
        configuration details within your Django settings. They are then
        injected into the view context to allow you to easily declare
        them within a template.
    """
    def get_context_data(self, **kwargs):
        """Overrides the view method to add Helcim.js details."""
        # Retrieve existing context
        context = super().get_context_data(**kwargs)

        # Add the Helcim.js configuration details
        context['helcim_js'] = SETTINGS['helcim_js']

        # Add a helper "test" input to allow declaration of test status
        # This input flags a transaction as a test and allows testing
        # in environments without SSL enabled (e.g. the Django
        # development server)
        for config in context['helcim_js']:
            if context['helcim_js'][config].get('test', False):
                test_input = mark_safe(
                    '<input id="test" type="hidden" value="1">'
                )
            else:
                test_input = ''

            context['helcim_js'][config]['test_input'] = test_input

        return context



# pylint: disable=missing-docstring, invalid-name
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('helcim', '0002_token_update_v2'),
    ]

    operations = [
        migrations.AddField(
            model_name='helcimtransaction',
            name='django_user',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='helcim_transactions',
                to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



"""Interface functions with the Helcim Commerce API.

These functions provide an agonstic interface with the Helcim Commerce
API and should work in any application.
"""
import requests
import xmltodict

from django.db import IntegrityError

from helcim import (
    conversions, exceptions as helcim_exceptions, mixins, models
)
from helcim.settings import SETTINGS


class BaseRequest(mixins.ResponseMixin):
    """Base class to handle validation and submission to Helcim API.

    Parameters:
        api_details (dict): Details to connect to Helcim API:

            - **url** (*str*): API URL.
            - **account_id** (*str*): Helcim account ID.
            - **token** (*str*): Helcim API token.
            - **terminal_id** (*str*): Helcim terminal ID.

        django_user (obj): The Django model for the requesting user.
        **kwargs (dict): Any additional transaction details.

    Keyword Arguments:
        amount (dec, optional): The amount for the transaction.
        currency (str, optional): The currency for the transaction.
        cc_name (str, optional): Name of the credit cardholder.
        cc_number (str, optional): 16 digit credit card number.
        cc_expiry (str, optional): 4 digit (MMYY) credit card
            expiry.
        cc_cvv (str, optional): 3 or 4 digit credit card CVV.
        cc_address (str, optional): Address of the credit
            cardholder.
        cc_postal_code (str, optional): Postal code/zip code of
            the credit cardholder.
        customer_code (str, optional): Helcim customer code.
        token (str, optional): 23 digit Helcim card token.
        token_f4l4 (str, optional): 8 digit "first four digits and
            last four digits" of the credit card number
        token_f4l4_skip (bool, optional): Whether to skip the F4L4
            verification.
        mag (string, optional): Non-encrypted credit card magnetic
            strip data.
        mag_enc (str, optional): Encrypted credit card magnestic
            strip data.
        mag_enc_serial_number (string, optional): Terminal serial
            number.
        order_number (str, optional): An assigned order number for
            the purchase.
        ecommerce (bool, optional): Whether this is an e-commerce
            transaction or not.
        comments (str, optional): Any additional comments with this
            transaction.
        billing_contact_name (str, optional): Billing address
            contact name.
        billing_business_name (str, optional): Billing address
            business name.
        billing_street_1 (str, optional): Billing street address
            (line 1).
        billing_street_2 (str, optional): Billing street address
            (line 2).
        billing_city (str, optional): Billing city.
        billing_province (str, optional): Billing province.
        billing_country (str, optional): Billing country.
        billing_postal_code (str, optional): Billing postal code.
        billing_phone (str, optional): Billing phone number.
        billing_fax (str, optional): Billing fax number
        billing_email (str, optional): Billing email.
        shipping_contact_name (str, optional): Shipping contact
                name.
        shipping_business_name (str, optional): Shipping business
            name.
        shipping_street_1 (str, optional): Shipping street address
            (line 1).
        shipping_street_2 (str, optional): Shipping street address
            (line 2).
        shipping_city (str, optional): Shipping city.
        shipping_province (str, optional): Shipping province.
        shipping_country (str, optional): Shipping country.
        shipping_postal_code (str, optional): Shipping postal code.
        shipping_phone (str, optional): Shipping phone number.
        shipping_fax (str, optional): Shipping fax number.
        shipping_email (str, optional): Shipping email address.
        amount_shipping (dec, optional): Shipping cost.
        amount_tax (dec, optional): Tax amount.
        shipping_method (str, optional): Method of shipping.
        tax_details (str, optional): Name for the tax (e.g. GST).
        test (bool, optional): Whether this is a test transaction or not.
    """

    def __init__(self, api_details=None, django_user=None, **kwargs):
        self.api = self.set_api_details(api_details)
        self.details = kwargs
        self.cleaned = {}
        self.response = {}
        self.redacted_response = {}
        self.django_user = django_user

    def set_api_details(self, details):
        """Sets the API details for this transaction.

            Will either return a dictionary of the API details from the
            provided details argument, or will look to the Django
            settings file.

            Parameters:
                details (dict): A dictionary of the API details.

            Returns:
                dict: The proper API details from the provided data.

            Raises:
                ImproperlyConfigured: A required API setting is not
                    found.
        """
        api_details = {
            'url': None,
            'account_id': None,
            'token': None,
            'terminal_id': None,
        }

        # Tries and retrieves settings form API details
        if details:
            api_details['url'] = details.get('url')
            api_details['account_id'] = details.get('account_id')
            api_details['token'] = details.get('token')
            api_details['terminal_id'] = details.get('terminal_id')

        # For any missing values, use values/defaults from settings
        if api_details['url'] is None:
            api_details['url'] = SETTINGS['api_url']

        if api_details['account_id'] is None:
            api_details['account_id'] = SETTINGS['account_id']

        if api_details['token'] is None:
            api_details['token'] = SETTINGS['api_token']

        if api_details['terminal_id'] is None:
            api_details['terminal_id'] = SETTINGS['terminal_id']

        return api_details

    def configure_test_transaction(self):
        """Adds test flag to post data if HELCIM_API_TEST is True.

            Method applies to the cleaned data (not the raw POST data).
            If the test flag is declared in both the POST data and
            Django settings file, the POST data takes precedence.
        """

        if 'test' not in self.cleaned and SETTINGS['api_test']:
            self.cleaned['test'] = SETTINGS['api_test']

    def process_error_response(self, response_message):
        """Returns error response with proper exception type.

            Parameters:
                response_message (str): The API error response message.

            Raises:
                PaymentError: Raised for any errors returned during a
                    purchase or refund transaction.
                HelcimError: Raised for any other unhandled errors.
        """
        exception_message = 'Helcim API request failed: {}'.format(
            response_message
        )

        if isinstance(self, Purchase):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Preauthorize):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Capture):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Refund):
            raise helcim_exceptions.RefundError(exception_message)

        if isinstance(self, Verification):
            raise helcim_exceptions.VerificationError(exception_message)

        raise helcim_exceptions.HelcimError(exception_message)

    def post(self, post_data=None):
        """Makes POST to Helcim API and updates response attribute.

        Parameters:
            post_data (dict): The parameters to pass with the POST request.

        Raises:
            ProcessingError: An error occurred connecting or
                communicating with Helcim API.
        """
        # Make the POST request
        try:
            response = requests.post(
                self.api['url'],
                data=post_data,
            )
        except requests.ConnectionError:
            raise helcim_exceptions.ProcessingError(
                'Unable to connect to Helcim API ({})'.format(self.api['url'])
            )

        # Catch any response errors in status code
        if response.status_code != 200:
            raise helcim_exceptions.ProcessingError(
                'Helcim API request failed with status code {}'.format(
                    response.status_code
                )
            )

        # Create the dictionary ('message' is the XML structure object)
        dict_response = xmltodict.parse(response.text)['message']

        # Catch any issues with the API response
        if dict_response['response'] == '0':
            self.process_error_response(dict_response['responseMessage'])

        # Return the response
        self.response = conversions.process_api_response(
            dict_response,
            post_data,
            response.text
        )

    def save_transaction(self, transaction_type):
        """Saves provided transaction data as Django model instance.

            Parameters:
                transaction_type (str): The type of transaction (``s``
                    for purchase (sale), ``p`` for pre-authorization,
                    ``c`` for capture, and ``r`` for refund).

            Returns:
                obj: A Django model instance of the saved data.

            Raises:
                DjangoError: Issue when attempting to save transaction
                    to database.
        """
        # Redacts data if not already done
        if not bool(self.redacted_response):
            self.redact_data()

        model_dictionary = self.create_model_arguments(transaction_type)

        try:
            saved_model = models.HelcimTransaction.objects.create(
                **model_dictionary
            )
        except IntegrityError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )
        except ValueError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )

        return saved_model

    def validate_fields(self):
        """Validates Helcim API request fields and coerces values."""
        self.cleaned = conversions.validate_request_fields(self.details)

class BaseCardTransaction(BaseRequest):
    """Base class for transactions involving credit card details."""
    def __init__(self, save_token=False, **kwargs):
        """Extends BaseRequest to include save_token and django_user.

            Parameters:
                save_token (bool): Whether the user has requested this
                    token to be saved or not.
        """
        super(BaseCardTransaction, self).__init__(**kwargs)
        self.save_token = self._determine_save_token_status(save_token)

    def determine_card_details(self):
        """Confirms valid payment details and updates self.cleaned.

        Cycles through the provided details to determine the most
        appropriate payment method. If multiple methods provided, only the
        first match is returned (token > customer code > CC number >
        encrypted magnetic strip > magnetic strip).

        Raises:
            ValueError: No valid payment details provided.
        """

        payment_fields = [
            'token', 'customer_code', 'token_f4l4', 'token_f4l4_skip',
            'cc_number', 'cc_expiry', 'cc_name', 'cc_cvv', 'cc_address',
            'cc_postal_code', 'mag_enc', 'mag_enc_serial_number', 'mag'
        ]

        if 'token' in self.cleaned and 'customer_code' in self.cleaned:
            # F4L4 required or it must be explicitly skipped
            f4l4_skip = self.cleaned.get('token_f4l4_skip', False)

            if f4l4_skip:
                payment_fields.remove('token')
                payment_fields.remove('customer_code')
                payment_fields.remove('token_f4l4_skip')

            if 'token_f4l4' in self.cleaned:
                payment_fields.remove('token')
                payment_fields.remove('customer_code')
                payment_fields.remove('token_f4l4')

        elif 'customer_code' in self.cleaned:
            payment_fields.remove('customer_code')

        elif 'cc_number' in self.cleaned and 'cc_expiry' in self.cleaned:
            payment_fields.remove('cc_number')
            payment_fields.remove('cc_expiry')

            # Add any additional CC details (as provided)
            if 'cc_name' in self.cleaned:
                payment_fields.remove('cc_name')

            if 'cc_cvv' in self.cleaned:
                payment_fields.remove('cc_cvv')

            if 'cc_address' in self.cleaned:
                payment_fields.remove('cc_address')

            if 'cc_postal_code' in self.cleaned:
                payment_fields.remove('cc_postal_code')

        elif (
                'mag_enc' in self.cleaned
                and 'mag_enc_serial_number' in self.cleaned
        ):
            payment_fields.remove('mag_enc')
            payment_fields.remove('mag_enc_serial_number')

        elif 'mag' in self.cleaned:
            payment_fields.remove('mag')

        # Raise error if no fields have been removed
        if len(payment_fields) == 13:
            raise ValueError('No valid payment details provided.')

        # Remove any other fields from self.cleaned
        for field in payment_fields:
            self.cleaned.pop(field, None)

class Purchase(BaseCardTransaction):
    """Makes a purchase request to Helcim Commerce API."""
    def process(self):
        """Makes a purchase request.

            Returns:
                tuple: The saved HelcimTransaction model
                    instance and HelcimToken model.
        """
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        purchase_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'purchase',
            }
        )
        self.post(purchase_data)

        purchase = self.save_transaction('s')
        token = self.save_token_to_vault()

        return purchase, token

class Preauthorize(BaseCardTransaction):
    """Makes a pre-authorization request to Helcim Commerce API."""
    def process(self):
        """Makes a pre-authorization request."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        preauth_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'preauth',
            }
        )
        self.post(preauth_data)

        preauth = self.save_transaction('p')
        token = self.save_token_to_vault()

        return preauth, token

class Refund(BaseCardTransaction):
    """Makes a refund request."""
    def process(self):
        """Makes a refund request to Helcim Commerce API."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        refund_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'refund',
            }
        )
        self.post(refund_data)

        refund = self.save_transaction('r')
        token = self.save_token_to_vault()

        return refund, token

class Verification(BaseCardTransaction):
    """Makes a verification request to Helcim Commerce API."""
    def process(self):
        """Makes a verification request to Helcim Commerce API."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        verification_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'verify',
            }
        )
        self.post(verification_data)

        verification = self.save_transaction('v')
        token = self.save_token_to_vault()

        return verification, token

class Capture(BaseRequest):
    """Makes a capture request (to complete a preauthorization)."""
    def validate_preauth_transaction(self):
        """Confirms that a preauth transaction ID was provided.

            Raises:
                PaymentError: Error when no transaction ID provided.
        """

        if 'transaction_id' not in self.details:
            raise helcim_exceptions.PaymentError(
                'Transaction ID must be provided with capture (force) request.'
            )

    def process(self):
        """Completes a capture request."""
        self.validate_fields()
        self.validate_preauth_transaction()
        self.configure_test_transaction()

        capture_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'capture',
            }
        )

        self.post(capture_data)
        capture = self.save_transaction('c')

        return capture

class HelcimJSResponse(mixins.ResponseMixin):
    """Class to handle Helcim.js Responses.

        This is a helper class that takes a Helcim.js response,
        handles any errors, converts the response into Python types
        and applies any redactions.

        Handling is more limited since all requests to the API are
        handled directly by Helcim.js. It is important to note that
        responses from Helcim.js are similar, but not identical to the
        basic Helcim API.

        Parameters:
            response (obj): the Helcim.js response POST data.
            save_token (bool): whether to save this to the
                ``django-helcim`` token vault.
            django_user (obj): a django user instance to use with
                ``django-helcim`` model instances.
    """
    def __init__(self, response, save_token=False, django_user=None):
        self.raw_response = response
        self.response = {}
        self.redacted_response = {}
        self.save_token = self._determine_save_token_status(save_token)
        self.django_user = django_user
        self.validated = False
        self.valid = False


    def is_valid(self):
        """Validates format is correct and notifies of any errors.

            Will also generate the redacted response at this point.

            Returns:
                boolean: True if valid without errors, otherwise false.
        """
        # Save the converted response
        self.response = conversions.process_helcim_js_response(
            self.raw_response
        )

        # Apply any redactions as needed
        self.redact_data()

        # Updates instance to note that validation was run and record result
        self.validated = True
        self.valid = self.response['transaction_success']

        return self.valid

    def _record_response(self, transaction_type):
        """Handles saving transaction and token for various response types.

            Will run last checks to ensure data is valid before saving to
            database.

            Parameters:
                transaction_type (str): the transaction_type to save the
                    transaction as: purchase/sale (``s``), preauthorization
                    (``p``), or verification (``v``).

            Returns:
                tuple: tuple of HelcimTransaction and HelcimToken instances.
                    HelcimToken will be None if token is not saved.
        """
        if self.validated is False:
            raise helcim_exceptions.HelcimError(
                'Must validate data with the .is_valid() method '
                ' before recording purchase.'
            )

        if self.valid is False:
            raise helcim_exceptions.HelcimError(
                'Response data was invalid - cannot record purchase data.'
            )

        transaction_instance = self.save_transaction(transaction_type)
        token_instance = self.save_token_to_vault()

        return transaction_instance, token_instance

    def record_purchase(self):
        """Saves validated purchase response to database.

            Returns:
                tuple: tuple of HelcimTransaction and HelcimToken instances.
                    HelcimToken will be None if token is not saved.
        """
        return self._record_response('s')

    def record_preauthorization(self):
        """Saves validated preauthorization response to database.

            Returns:
                tuple of the HelcimTransaction and HelcimToken instances.
                    HelcimToken will be None if token is not saved.
        """
        return self._record_response('p')

    def record_verification(self):
        """Saves validated verification response to database.

            Returns:
                tuple of the HelcimTransaction and HelcimToken instances.
                    HelcimToken will be None if token is not saved.
        """
        return self._record_response('v')

def retrieve_token_details(token_id, django_user=None, customer_code=None):
    """Takes a HelcimToken ID and maps details to dictionary."""
    # Final validation to ensure token exists & belongs to proper user
    try:
        token_instance = models.HelcimToken.objects.get(
            id=token_id,
            django_user=django_user,
            customer_code=customer_code
        )
    except models.HelcimToken.DoesNotExist:
        raise helcim_exceptions.ProcessingError(
            'Unable to retrieve token details for specified customer.'
        )

    # Validation passed - map model to the dictionary
    return {
        'token': token_instance.token,
        'token_f4l4': token_instance.token_f4l4,
        'django_user': token_instance.django_user,
        'customer_code': token_instance.customer_code,
    }

def retrieve_saved_tokens(django_user=None, customer_code=None):
    """Returns list of tokens for specified customer.

        Parameters:
            django_user (obj): A Django user model instance.
            customer_code (str): A Helcim customer code.

        Returns:
            obj: A queryset of the retrieved tokens
    """
    if django_user and customer_code:
        return models.HelcimToken.objects.filter(
            django_user=django_user, customer_code=customer_code
        )

    if django_user:
        return models.HelcimToken.objects.filter(django_user=django_user)

    if customer_code:
        return models.HelcimToken.objects.filter(customer_code=customer_code)

    return models.HelcimToken.objects.none()



"""Models for the django-helcim application."""
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.contrib.staticfiles import finders
from django.db import models


class HelcimTransaction(models.Model):
    """Details of a single Helcim transaction."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    raw_request = models.CharField(
        blank=True,
        help_text='The raw request used for this transaction',
        max_length=1024,
        null=True,
    )
    raw_response = models.CharField(
        blank=True,
        help_text='The raw response returned for this transaction',
        max_length=1024,
        null=True,
    )
    transaction_success = models.BooleanField(
        help_text='Whether the transaction was successful or not',
    )
    response_message = models.CharField(
        blank=True,
        help_text='The response message with the API call',
        max_length=256,
        null=True,
    )
    notice = models.CharField(
        blank=True,
        help_text='Any error or warning messages from Helcim',
        max_length=128,
        null=True,
    )
    date_response = models.DateTimeField(
        help_text='The date and time of the API response',
    )
    date_created = models.DateTimeField(
        auto_now_add=True,
        help_text='Date and time this transaction was recorded in database',
    )
    TRANSACTION_TYPES = (
        ('s', 'purchase (sale)'),
        ('p', 'pre-authorization'),
        ('c', 'capture'),
        ('r', 'refund'),
        ('v', 'verify'),
    )
    transaction_type = models.CharField(
        choices=TRANSACTION_TYPES,
        help_text='The type of transaction',
        max_length=1,
    )
    transaction_id = models.PositiveIntegerField(
        blank=True,
        help_text='The Helcim Commerce transaction ID',
        null=True,
    )
    amount = models.DecimalField(
        blank=True,
        decimal_places=2,
        help_text='The transaction amount',
        max_digits=12,
        null=True,
    )
    currency = models.CharField(
        blank=True,
        help_text='The transaction currency',
        max_length=8,
        null=True,
    )
    cc_name = models.CharField(
        blank=True,
        help_text='The cardholder name',
        max_length=256,
        null=True,
    )
    cc_number = models.CharField(
        blank=True,
        help_text='The first four and last 4 digits of the credit card number',
        max_length=16,
        null=True,
    )
    cc_expiry = models.DateField(
        blank=True,
        help_text='The credit card expiry date',
        null=True,
    )
    cc_type = models.CharField(
        blank=True,
        help_text='The credit card type',
        max_length=32,
        null=True,
    )
    token = models.CharField(
        blank=True,
        help_text='The Helcim generated and stored credit card token',
        max_length=23,
        null=True,
    )
    token_f4l4 = models.CharField(
        blank=True,
        help_text='The first and last 4 digits of the credit card number',
        max_length=8,
        null=True,
    )
    avs_response = models.CharField(
        blank=True,
        help_text='The address verification response',
        max_length=1,
        null=True,
    )
    cvv_response = models.CharField(
        blank=True,
        help_text='The CVV verification response',
        max_length=1,
        null=True,
    )
    approval_code = models.CharField(
        blank=True,
        help_text='The transaction approval code',
        max_length=16,
        null=True,
    )
    order_number = models.CharField(
        blank=True,
        help_text='The Helcim order number',
        max_length=16,
        null=True,
    )
    customer_code = models.CharField(
        blank=True,
        help_text='The Helcim customer code',
        max_length=16,
        null=True,
    )
    django_user = models.ForeignKey(
        get_user_model(),
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name='helcim_transactions',
    )

    class Meta:
        ordering = ('-date_response',)
        permissions = (
            (
                'helcim_transactions',
                'Can view and interact with Helcim transactions.'
            ),
        )

    def __str__(self):
        string_parts = [
            str(self.date_response),
            self.transaction_type,
        ]

        return ' - '.join(string_parts)

    @property
    def can_be_captured(self):
        """Check if this transaction can be captured."""
        return bool(
            self.transaction_success and self.transaction_type == 'p'
        )

    @property
    def can_be_refunded(self):
        """Check if this transaction can be refunded."""
        return all([
            self.transaction_success,
            (self.transaction_type == 's' or self.transaction_type == 'c'),
            (self.amount if self.amount else 0) > 0,
        ])

class HelcimToken(models.Model):
    """A Helcim card token."""
    id = models.UUIDField(
        default=uuid4,
        editable=False,
        primary_key=True,
        verbose_name='ID',
    )
    token = models.CharField(
        help_text='The Helcim card token number',
        max_length=23,
    )
    token_f4l4 = models.CharField(
        help_text='The first & last four digits of the credit card number',
        max_length=8
    )
    cc_name = models.CharField(
        blank=True,
        help_text='The cardholder name',
        max_length=256,
        null=True,
    )
    cc_expiry = models.DateField(
        blank=True,
        help_text='The credit card expiry date',
        null=True,
    )
    cc_type = models.CharField(
        blank=True,
        help_text='The credit card type',
        max_length=32,
        null=True,
    )
    date_added = models.DateTimeField(
        auto_now_add=True,
        help_text='Date and time this token was added to database',
    )
    customer_code = models.CharField(
        blank=True,
        help_text='The Helcim customer code',
        max_length=16,
        null=True,
    )
    django_user = models.ForeignKey(
        get_user_model(),
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name='helcim_tokens',
    )

    class Meta:
        permissions = (
            (
                'helcim_tokens',
                'Can view and interact with Helcim tokens.'
            ),
        )
        unique_together = (
            'token', 'token_f4l4', 'customer_code', 'django_user'
        )

    @property
    def display_as_card_number(self):
        """Displays token_f4l4 as a 16 character credit card number."""
        return '{}********{}'.format(self.token_f4l4[:4], self.token_f4l4[-4:])

    @property
    def get_credit_card_png(self):
        """Returns a path to a credit card .png for this token."""
        # Look for an image for this credit card
        if self.cc_type:
            image_path = 'helcim/{}.png'.format(self.cc_type.lower())

            # If found, return the path to the static file image
            if finders.find(image_path):
                return image_path

        # Return placeholder image
        return 'helcim/placeholder.png'

    @property
    def get_credit_card_svg(self):
        """Returns a path to a credit card .svg for this token."""
        # Look for an image for this credit card
        if self.cc_type:
            image_path = 'helcim/{}.svg'.format(self.cc_type.lower())

            # If found, return the path to the static file image
            if finders.find(image_path):
                return image_path

        # Return placeholder image
        return 'helcim/placeholder.svg'



"""Views for Helcim Commerce API transactions."""
from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.urls import reverse, reverse_lazy
from django.views import generic

from helcim import exceptions, models, gateway


class TransactionListView(PermissionRequiredMixin, generic.ListView):
    """List of all transactions submitted made by django-helcim."""
    model = models.HelcimTransaction
    permission_required = 'helcim.helcim_transactions'
    raise_exception = True
    context_object_name = 'transactions'
    template_name = 'helcim/transaction_list.html'

class TransactionDetailView(PermissionRequiredMixin, generic.DetailView):
    """Details of a specific transaction made by django-helcim."""
    model = models.HelcimTransaction
    pk_url_kwarg = 'transaction_id'
    permission_required = 'helcim.helcim_transactions'
    raise_exception = True
    context_object_name = 'transaction'
    template_name = 'helcim/transaction_detail.html'

    def get_context_data(self, **kwargs):
        context = super(TransactionDetailView, self).get_context_data(**kwargs)

        # Check which actions are enabled
        settings = gateway.SETTINGS
        context['capture_enabled'] = settings['enable_transaction_capture']
        context['refund_enabled'] = settings['enable_transaction_refund']

        return context

    def post(self, request, *args, **kwargs): # pylint: disable=unused-argument
        """Handles any capture and refund requests."""
        # Determines which action was submitted
        action = request.POST.get('action', None)

        if action == 'refund':
            if gateway.SETTINGS['enable_transaction_refund']:
                return self.refund()

            transaction = self.get_object()
            messages.error(self.request, 'Transactions cannot be refunded')

            return HttpResponseRedirect(
                reverse(
                    'helcim_transaction_detail',
                    kwargs={'transaction_id': transaction.id}
                )
            )

        if action == 'capture':
            if gateway.SETTINGS['enable_transaction_capture']:
                return self.capture()

            transaction = self.get_object()
            messages.error(self.request, 'Transactions cannot be captured')

            return HttpResponseRedirect(
                reverse(
                    'helcim_transaction_detail',
                    kwargs={'transaction_id': transaction.id}
                )
            )

        # Action not found - return bad request
        return HttpResponseBadRequest('Unrecognized transaction action')

    def refund(self):
        """Processes a refund transaction."""
        transaction = self.get_object()

        refund = gateway.Refund({
            'transaction_type': 'r',
            'amount': transaction.amount,
            'token': transaction.token,
            'token_f4l4': transaction.token_f4l4,
            'customer_code': transaction.customer_code,
        })

        try:
            refund.process()
        except exceptions.RefundError:
            messages.error(self.request, 'Unable to refund transaction')
        else:
            messages.success(self.request, 'Transaction refunded')

        return HttpResponseRedirect(
            reverse(
                'helcim_transaction_detail',
                kwargs={'transaction_id': transaction.id}
            )
        )

    def capture(self):
        """Processes a capture transaction."""
        transaction = self.get_object()

        capture = gateway.Capture({
            'transaction_type': 'c',
            'transaction_id': transaction.transaction_id,
        })

        try:
            capture.process()
        except exceptions.PaymentError:
            messages.error(self.request, 'Unable to capture transaction')
        else:
            messages.success(self.request, 'Transaction captured')

        return HttpResponseRedirect(
            reverse(
                'helcim_transaction_detail',
                kwargs={'transaction_id': transaction.id}
            )
        )

class TokenListView(PermissionRequiredMixin, generic.ListView):
    """List of all transactions submitted made by django-helcim."""
    model = models.HelcimToken
    permission_required = 'helcim.helcim_tokens'
    raise_exception = True
    context_object_name = 'tokens'
    template_name = 'helcim/token_list.html'

class TokenDeleteView(PermissionRequiredMixin, generic.DeleteView):
    """Allows deletion of a Helcim API token."""
    model = models.HelcimToken
    permission_required = 'helcim.helcim_tokens'
    raise_exception = True
    pk_url_kwarg = 'token_id'
    context_object_name = 'token'
    success_message = 'Token successfully deleted.'
    success_url = reverse_lazy('helcim_token_list')
    template_name = 'helcim/token_delete.html'

    def delete(self, request, *args, **kwargs):
        """Override delete to allow success message to be added."""
        messages.success(self.request, self.success_message)
        return super(TokenDeleteView, self).delete(request, *args, **kwargs)



"""Admin settings for Helcim Commerce API transactions."""
from django.contrib import admin

from helcim.gateway import SETTINGS
from helcim.models import HelcimTransaction, HelcimToken

class HelcimTransactionAdmin(admin.ModelAdmin):
    """Admin class for the HelcimTransaction model."""
    MODEL_FIELDS = [
        'raw_request',
        'raw_response',
        'transaction_success',
        'response_message',
        'notice',
        'date_response',
        'date_created',
        'transaction_type',
        'transaction_id',
        'amount',
        'currency',
        'cc_name',
        'cc_number',
        'cc_expiry',
        'cc_type',
        'token',
        'token_f4l4',
        'avs_response',
        'cvv_response',
        'approval_code',
        'order_number',
        'customer_code',
        'django_user',
    ]

    fields = MODEL_FIELDS

    readonly_fields = MODEL_FIELDS

    list_display = [
        'transaction_type',
        'date_response',
        'transaction_success',
        'amount',
        'customer_code',
        'django_user'
    ]

class HelcimTokenAdmin(admin.ModelAdmin):
    """Admin class for the HelcimToken model."""
    MODEL_FIELDS = [
        'token',
        'token_f4l4',
        'cc_name',
        'cc_expiry',
        'date_added',
        'customer_code',
        'django_user',
        'cc_type',
    ]

    fields = MODEL_FIELDS

    readonly_fields = MODEL_FIELDS

    list_display = [
        'date_added',
        'django_user',
        'customer_code',
        'token',
        'token_f4l4',
        'cc_name',
        'cc_expiry',
        'cc_type',
    ]

# Only register admin models if enabled in settings
if SETTINGS['enable_admin']:
    admin.site.register(HelcimTransaction, HelcimTransactionAdmin)
    admin.site.register(HelcimToken, HelcimTokenAdmin)



# pylint: disable=missing-docstring, line-too-long, invalid-name

from __future__ import unicode_literals
from uuid import uuid4

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='HelcimTransaction',
            fields=[
                ('id', models.UUIDField(default=uuid4, editable=False, primary_key=True, verbose_name='ID')),
                ('raw_request', models.CharField(blank=True, help_text='The raw request used for this transaction', max_length=1024, null=True)),
                ('raw_response', models.CharField(blank=True, help_text='The raw response returned for this transaction', max_length=1024, null=True)),
                ('transaction_success', models.BooleanField(help_text='Whether the transaction was successful or not')),
                ('response_message', models.CharField(blank=True, help_text='The response message with the API call', max_length=256, null=True)),
                ('notice', models.CharField(blank=True, help_text='Any error or warning messages from Helcim', max_length=128, null=True)),
                ('date_response', models.DateTimeField(help_text='The date and time of the API response')),
                ('date_created', models.DateTimeField(auto_now_add=True, help_text='Date and time this transaction was recorded in database')),
                ('transaction_type', models.CharField(choices=[('s', 'purchase (sale)'), ('p', 'pre-authorization'), ('c', 'capture'), ('r', 'refund')], help_text='The type of transaction', max_length=1)),
                ('transaction_id', models.PositiveIntegerField(blank=True, help_text='The Helcim Commerce transaction ID', null=True)),
                ('amount', models.DecimalField(blank=True, decimal_places=2, help_text='The transaction amount', max_digits=12, null=True)),
                ('currency', models.CharField(blank=True, help_text='The transaction currency', max_length=8, null=True)),
                ('cc_name', models.CharField(blank=True, help_text='The cardholder name', max_length=256, null=True)),
                ('cc_number', models.CharField(blank=True, help_text='The first four and last 4 digits of the credit card number', max_length=16, null=True)),
                ('cc_expiry', models.DateField(blank=True, help_text='The credit card expiry date', null=True)),
                ('cc_type', models.CharField(blank=True, help_text='The credit card type', max_length=32, null=True)),
                ('token', models.CharField(blank=True, help_text='The Helcim generated and stored credit card token', max_length=23, null=True)),
                ('token_f4l4', models.CharField(blank=True, help_text='The first and last 4 digits of the credit card number', max_length=8, null=True)),
                ('avs_response', models.CharField(blank=True, help_text='The address verification response', max_length=1, null=True)),
                ('cvv_response', models.CharField(blank=True, help_text='The CVV verification response', max_length=1, null=True)),
                ('approval_code', models.CharField(blank=True, help_text='The transaction approval code', max_length=16, null=True)),
                ('order_number', models.CharField(blank=True, help_text='The Helcim order number', max_length=16, null=True)),
                ('customer_code', models.CharField(blank=True, help_text='The Helcim customer code', max_length=16, null=True)),
            ],
            options={
                'ordering': ('-date_response',),
                'permissions': (('helcim_transactions', 'Can view and interact with Helcim transactions.'),),
            },
        ),
        migrations.CreateModel(
            name='HelcimToken',
            fields=[
                ('id', models.UUIDField(default=uuid4, editable=False, primary_key=True, verbose_name='ID')),
                ('token', models.CharField(help_text='The Helcim card token number', max_length=23)),
                ('token_f4l4', models.CharField(help_text='The first & last four digits of the credit card number', max_length=8)),
                ('cc_type', models.CharField(blank=True, help_text='The credit card type', max_length=32, null=True)),
                ('date_added', models.DateTimeField(auto_now_add=True, help_text='Date and time this token was added to database')),
                ('customer_code', models.CharField(blank=True, help_text='The Helcim customer code', max_length=16, null=True)),
                ('django_user', models.ForeignKey(blank=True, null=True, on_delete=models.deletion.CASCADE, related_name='tokens', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'permissions': (('helcim_tokens', 'Can view and interact with Helcim tokens.'),),
                'unique_together': {('token', 'token_f4l4', 'customer_code', 'django_user')},
            },
        )
    ]



# pylint: disable=missing-docstring, invalid-name
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('helcim', '0003_add_user_to_transaction'),
    ]

    operations = [
        migrations.AlterField(
            model_name='helcimtransaction',
            name='transaction_type',
            field=models.CharField(
                choices=[
                    ('s', 'purchase (sale)'),
                    ('p', 'pre-authorization'),
                    ('c', 'capture'),
                    ('r', 'refund'),
                    ('v', 'verify')
                ],
                help_text='The type of transaction',
                max_length=1,
            ),
        ),
    ]



"""URLs to integrate with django-oscar."""
# pylint: disable=line-too-long
from django.conf.urls import url

from helcim import gateway, views

urlpatterns = [
    url(
        r'^transactions/$',
        views.TransactionListView.as_view(),
        name='helcim_transaction_list'
    ),
    url(
        r'^transactions/(?P<transaction_id>[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12})/$',
        views.TransactionDetailView.as_view(),
        name='helcim_transaction_detail'
    ),
]

# Only add these views if token vault is enabled
if gateway.SETTINGS['enable_token_vault']:
    urlpatterns += [
        url(
            r'^tokens/$',
            views.TokenListView.as_view(),
            name='helcim_token_list'
        ),
        url(
            r'^tokens/(?P<token_id>[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[89ab][a-f0-9]{3}-?[a-f0-9]{12})/$',
            views.TokenDeleteView.as_view(),
            name='helcim_token_delete'
        ),
    ]



"""Custom exceptions for django-helcim."""

class HelcimError(Exception):
    """Base exception for all package exceptions."""
    pass

class ProcessingError(HelcimError):
    """Exception to handle system and API connection errors."""
    pass

class PaymentError(HelcimError):
    """Exception to handle payments, pre-auths, and captures."""
    pass

class RefundError(HelcimError):
    """Exception to handle refund errors."""
    pass

class VerificationError(HelcimError):
    """Exception to handle errors during credit care verification."""
    pass

class DjangoError(HelcimError):
    """Exception for issues relating to Django interface."""
    pass
